@@ -12,6 +12,7 @@ package Net::SIP::Endpoint::Context;
 use fields (
 
 	# ===== can be set with new()
+	'method',  # initiated by which method
 	'from',    # from where
 	'to',      # to where
 	'auth',    # [ user,pass ] or { realm1 => [ user1,pass1 ], realm2 => [ user2,pass2 ],... }
@@ -54,7 +55,7 @@ use Net::SIP::Util ':all';
 #   @args: either single \%args (hash-ref) or %args (hash) with at least
 #     values for from and to
 #     callid,cseq will be generated if not given
-#     routes will default to [] and usually set from record-route header
+#     routes will default to undef and usually set from record-route header
 #     in response packets
 # Returns: $self
 ############################################################################
@@ -65,7 +66,6 @@ sub new {
 	%$self = %args;
 	$self->{callid} ||= md5_hex( time(), rand(2**32) );
 	$self->{cseq} ||= 0;
-	$self->{route} ||= [];
 	$self->{_transactions} = [];
 	$self->{_cseq_incoming} = 0;
 
@@ -159,6 +159,7 @@ sub new_request {
 		# already a request object
 		$request = $method;
 		$method = $request->method;
+
 	} else {
 
 		# increase cseq unless its explicitly specified
@@ -187,16 +188,18 @@ sub new_request {
 				from => $from,
 				to => $to,
 				$self->{contact} ? ( contact => $self->{contact} ):(),
-				route => $self->{route},
 				cseq => "$cseq $method",
 				'call-id' => $self->{callid},
 				'max-forwards' => 70,
-				%args
+				%args,
 			},
 			$body
-		)
+		);
 	}
 
+	# overwrite any route header in request if we already learned a route
+	$request->set_header( route => $self->{route} ) if $self->{route};
+
 	# create new transaction
 	my %trans = (
 		tid      => $request->tid,
@@ -234,12 +237,14 @@ sub set_callback {
 sub request_delivery_done {
 	my Net::SIP::Endpoint::Context $self = shift;
 	my ($endpoint,$tid,$error) = @_;
+	return if ! $error; # notify of success once I get response
+
 	my $trans = $self->{_transactions};
 	my @ntrans;
 	foreach my $tr (@$trans) {
 		if ( $tr->{tid} eq $tid ) {
 			$self->{_transactions} = \@ntrans;
-			if ( $error && ( my $cb = $tr->{callback} )) {
+			if ( my $cb = $tr->{callback} ) {
 				# permanently failed
 				invoke_callback( $cb, $self,$endpoint,$error );
 			}
@@ -315,6 +320,10 @@ sub handle_response {
 			$endpoint->close_context( $self ) if $method eq 'BYE';
 		}
 		return;
+	} elsif ( $self->{method} ne 'INVITE' and 
+		($code>=200 and $code<300 or $code>=400 and $code != 401 and $code!= 407)) {
+		# final response in non-dialog (only INVITE can create dialog)
+		$endpoint->close_context($self);
 	}
 
 	# for 300-699 an ACK must be created (RFC3261, 17.1.1.2)
@@ -343,22 +352,26 @@ sub handle_response {
 	} elsif ( $code =~m{^2\d\d} ) {
 		# 2xx OK
 
-		# extract route information
-		if ( my @route = $response->get_header( 'record-route' )) {
-			$self->{route} = [ reverse @route ];
-		}
-
 		if ( $method eq 'INVITE' ) {
 			# is response to INVITE, create ACK
 			# and propagate to upper layer
 			my $req = $tr->{request};
 
+			# extract route information on INVIE, but not on re-INVITE
+			# we assume, that it is a re-INVITE, if we have a remote_contact
+			# already
+			if ( ! $self->{remote_contact}
+				and my @route = $response->get_header( 'record-route' )) {
+				$self->{route} = [ reverse @route ];
+			}
+
 			# 12.1.2 - set URI for dialog to contact given in response which
 			# establishes the dialog
 			if ( my $contact = $response->get_header( 'contact' )) {
 				$contact = $1 if $contact =~m{<(\w+:[^>\s]+)>};
 				$self->{remote_contact} = $contact;
 				$req->set_uri( $contact );
+
 			}
 
 			# use to-tag from this request to update 'to'
@@ -415,7 +428,7 @@ sub handle_response {
 	} elsif ( $code == 305 ) {
 		# 21.3.4 305 use proxy
 		# set proxy as the first route and insert request again
-		my $route = $self->{route};
+		my $route = $self->{route} ||= [];
 		unshift @$route,$response->get_header( 'contact' );
 		( my $r = $tr->{request} )->set_header( route => $route );
 		$r->set_cseq( ++$self->{cseq} );
@@ -454,12 +467,15 @@ sub handle_request {
 	}
 
 	# check with last request in transaction
+	my $ctx_is_new;
 	if ( my $trans = $self->{_last_transreq} ) {
 		my $last_cseq = $trans->cseq;
 		if ( $last_cseq eq $cseq ) {
 			DEBUG( 10,"retransmit of last request. DROP" );
 			return;
 		}
+	} else {
+		$ctx_is_new = 1;
 	}
 	$self->{_last_transreq} = $request;
 
@@ -487,7 +503,9 @@ sub handle_request {
 	my @arg = ($endpoint,$self);
 
 	# extract route information for future requests to the UAC (re-invites)
-	if ( my @route = $request->get_header( 'record-route' )) {
+	# only for INVITE (rfc3261,12.1.1)
+	if ( $ctx_is_new and $method eq 'INVITE' and 
+		my @route = $request->get_header( 'record-route' )) {
 		$self->{route} = \@route;
 	}
 
